// CVLOAD.C: CodeView loader

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#include "cvload.h"

// limits:
#define MAX_SYMBOL_NAME   255
#define MAX_SOURCE_FILE   100
#define MAX_SEGMENT       100

// debug code:
#define ERR(msg) { fprintf(stderr,"CVLOAD ERROR: " msg "\n"); return 0; }
//#define ERR(msg) return 0

// cvload structures:
typedef struct {
  unsigned ofs;     // linear offset
  unsigned short n; // line number
} tLine;

typedef struct {
  char *name;       // name of source module
  unsigned from;    // from linear address
  unsigned to;      // to linear address (excluding)
  int nrLine;       // # of line infos
  tLine *line;      // line infos
} tSource;

typedef struct {
  unsigned from;    // from linear address
  unsigned to;      // to linear address (excluding)
  char *name;       // name of proc entry
  tSource *src;     // matching source file (NULL possible)
} tProc;

// codeview defines and structures:
#define sstSymbols      0x124
#define sstAlignSym     0x125
#define sstSrcModule    0x127

#define S_LPROC32       0x0204
#define S_GPROC32       0x0205
#define S_LPROC32_VC5   0x100a
#define S_GPROC32_VC5   0x100b

typedef struct {
  unsigned short cbDirHeader;
  unsigned short cbDirEntry;
  unsigned cDir;
  unsigned lfoNextDir;
  unsigned flags;
} tDirHeader;

typedef struct {
  unsigned short subsection;
  unsigned short iMod;
  unsigned lfo;
  unsigned cb;
} tDirEntry;

typedef struct {
  unsigned pParent;
  unsigned pEnd;
  unsigned pNext;
  unsigned procLen;
  unsigned debugStart;
  unsigned debugEnd;
  unsigned ofs;
  unsigned short seg;
  unsigned short proctype;
  unsigned char flags;
  char name[256];
} tProcEntry;

typedef struct {
  unsigned pParent;
  unsigned pEnd;
  unsigned pNext;
  unsigned procLen;
  unsigned debugStart;
  unsigned debugEnd;
  unsigned proctype;
  unsigned ofs;
  unsigned short seg;
  unsigned char flags;
  char name[256];
} tProcEntryVC5;

typedef struct {
  unsigned short cFile;
  unsigned short cSeg;
} tSourceHeader;

typedef struct {
  unsigned start;
  unsigned end;
} tStartEnd;

typedef struct {
  unsigned short cSeg;
  unsigned short pad;
} tTable;

typedef struct {
  unsigned short Seg;
  unsigned short cPair;
} tLineHeader;

// local variables:
static FILE *f;           // current executable
static unsigned baseOfs;  // beginning of debug information
static unsigned dirStart; // start offset of directory
static char *strBuf;      // string buffer
static int curStrBuf;     // current index into string buffer
static tLine *line;       // line buffer
static int curLine;       // current index into line buffer
static tSource *src;      // source modules
static int nrSrc;         // # source modules
static tProc *proc;       // proc entries
static int nrProc;        // # proc entries
static int sizeString=0;  // string table
static int sizeProc=0;    // proc entry table
static int sizeSource=0;  // source file table
static int sizeLine=0;    // line number information

static int calcNeededMemory(void)
// calculates size of memory to be allocated and allocates
// structures
{
  tDirHeader hdr;
  unsigned k;

  // read header
  fseek(f,dirStart,SEEK_SET);
  fread(&hdr,sizeof(hdr),1,f);

  // sanity checks
  if (hdr.cbDirHeader!=sizeof(tDirHeader)) ERR("cbDirHeader!=tDirHeader");
  if (hdr.cbDirEntry!=sizeof(tDirEntry)) ERR("cbDirEntry!=tDirEntry");

  // walk directory
  for (k=0;k<hdr.cDir;k++) 
  {
    tDirEntry e;
    fread(&e,sizeof(e),1,f);

    switch(e.subsection) 
    {
      // symbol lists
      case sstSymbols:
      case sstAlignSym:
        {
          unsigned id;
          unsigned save=ftell(f);
          fseek(f,e.lfo+baseOfs,SEEK_SET);

          // check ID
          fread(&id,4,1,f); e.cb-=4;
          if (id!=1&&id!=2) ERR("id!=1&&id!=2");

          // walk entries
          while (e.cb) 
          {
            // get header
            unsigned short li[2];
            fread(li,4,1,f);
            if (li[0]+(unsigned)2>e.cb) ERR("li[0]+2>e.cb");
            e.cb-=li[0]+2;

            switch(li[1])
            {
              case S_LPROC32:
              case S_GPROC32:
                {
                  tProcEntry pe;
                  fread(&pe,1,li[0]-2,f);
                  if (pe.seg==1) 
                  {
                    sizeProc++;
                    sizeString+=*pe.name+1;
                  }
                }
                continue;
              case S_LPROC32_VC5:
              case S_GPROC32_VC5:
                {
                  tProcEntryVC5 pe;
                  fread(&pe,1,li[0]-2,f);
                  if (pe.seg==1) 
                  {
                    sizeProc++;
                    sizeString+=*pe.name+1;
                  }
                }
                continue;
            }            
      
            fseek(f,li[0]-2,SEEK_CUR); // skip entry
          }

          fseek(f,save,SEEK_SET);
        }
        break;
      // source modules/line numbers
      case sstSrcModule:
        {
          tSourceHeader hdr;
          unsigned baseSrcFile[MAX_SOURCE_FILE];
          unsigned i;
          unsigned save=ftell(f);

          fseek(f,e.lfo+baseOfs,SEEK_SET);
          fread(&hdr,1,sizeof(hdr),f);
          if (hdr.cFile>MAX_SOURCE_FILE) ERR("hdr.cFile>MAX_SOURCE_FILE");
          if (hdr.cSeg>MAX_SEGMENT) ERR("hdr.cSeg>MAX_SEGMENT");
          sizeSource+=hdr.cFile;

          fread(baseSrcFile,4,hdr.cFile,f);

          for (i=0;i<hdr.cFile;i++) 
          {
            tTable table;
            unsigned baseSrcLin[MAX_SEGMENT];
            unsigned char nameLen;
            unsigned j;

            fseek(f,e.lfo+baseOfs+baseSrcFile[i],SEEK_SET);
            
            fread(&table,1,sizeof(table),f);
            if (table.cSeg>MAX_SEGMENT) ERR("table.cSeg>MAX_SEGMENT");

            fread(baseSrcLin,4,table.cSeg,f);
            
            fseek(f,8*table.cSeg,SEEK_CUR);
            fread(&nameLen,1,1,f);
            sizeString+=nameLen+1;

            for (j=0;j<table.cSeg;j++) 
            {
              tLineHeader hdr;

              fseek(f,e.lfo+baseOfs+baseSrcLin[j],SEEK_SET);
              
              fread(&hdr,sizeof(hdr),1,f);
              if (hdr.Seg==1) sizeLine+=hdr.cPair;
            }
          }
          fseek(f,save,SEEK_SET);
        }
        break;
      }
    }

  // ok, now allocate structures
  strBuf=(char *)malloc(sizeString);
  if (!strBuf&&sizeString) ERR("strBuf");
  memset(strBuf,0,sizeString);
  line=(tLine *)malloc(sizeLine*sizeof(tLine));
  if (!line&&sizeLine) ERR("line");
  src=(tSource *)malloc(sizeSource*sizeof(tSource));
  if (!src&&sizeSource) ERR("src");
  proc=(tProc *)malloc(sizeProc*sizeof(tProc));
  if (!proc&&sizeProc) ERR("proc");

  // clear counter
  curStrBuf=0;
  curLine=0;
  nrSrc=0;
  nrProc=0;

  return 1;
}

static int loadData(void)
// loads codeview data into memory
{
  tDirHeader hdr;
  unsigned k;

  // read header
  fseek(f,dirStart,SEEK_SET);
  fread(&hdr,sizeof(hdr),1,f);

  // walk directory
  for (k=0;k<hdr.cDir;k++) 
  {
    tDirEntry e;
    fread(&e,sizeof(e),1,f);

    switch(e.subsection) 
    {
      // symbol lists
      case sstSymbols:
      case sstAlignSym:
        {
          unsigned save=ftell(f);
          fseek(f,e.lfo+baseOfs+4,SEEK_SET); // skipping ID
          e.cb-=4;

          // walk entries
          while (e.cb) 
          {
            // get header
            unsigned short li[2];
            fread(li,4,1,f);
            if (li[0]+(unsigned)2>e.cb) return 0;
            e.cb-=li[0]+2;

            switch(li[1])
            {
              case S_LPROC32:
              case S_GPROC32:
                {
                  tProcEntry pe;
                  fread(&pe,1,li[0]-2,f);
                  if (pe.seg==1) 
                  {
                    // add proc entry
                    assert(nrProc<sizeProc);
                    proc[nrProc].from=pe.ofs;
                    proc[nrProc].to=pe.ofs+pe.procLen;
                    proc[nrProc].name=strBuf+curStrBuf;
                    proc[nrProc].src=NULL; // seek matching source files later
                    nrProc++;

                    // add string
                    memcpy(strBuf+curStrBuf,pe.name+1,*pe.name);
                    curStrBuf+=*pe.name+1;
                  }
                }
                continue;
              case S_LPROC32_VC5:
              case S_GPROC32_VC5:
                {
                  tProcEntryVC5 pe;
                  fread(&pe,1,li[0]-2,f);
                  if (pe.seg==1) 
                  {
                    // add proc entry
                    assert(nrProc<sizeProc);
                    proc[nrProc].from=pe.ofs;
                    proc[nrProc].to=pe.ofs+pe.procLen;
                    proc[nrProc].name=strBuf+curStrBuf;
                    proc[nrProc].src=NULL; // seek matching source files later
                    nrProc++;

                    // add string
                    memcpy(strBuf+curStrBuf,pe.name+1,*pe.name);
                    curStrBuf+=*pe.name+1;
                  }
                }
                continue;
            }            
      
            fseek(f,li[0]-2,SEEK_CUR); // skip entry
          }

          fseek(f,save,SEEK_SET);
        }
        break;
      // source modules/line numbers
      case sstSrcModule:
        {
          tSourceHeader hdr;
          unsigned baseSrcFile[MAX_SOURCE_FILE];
          unsigned short seg[MAX_SEGMENT];
          unsigned i;
          unsigned save=ftell(f);

          fseek(f,e.lfo+baseOfs,SEEK_SET);
          fread(&hdr,1,sizeof(hdr),f);
          fread(baseSrcFile,4,hdr.cFile,f);
          fseek(f,8*hdr.cSeg,SEEK_CUR);
          fread(seg,2,hdr.cSeg,f);

          for (i=0;i<hdr.cFile;i++) 
          {
            tTable table;
            unsigned baseSrcLin[MAX_SEGMENT];
            tStartEnd se[MAX_SEGMENT];
            unsigned char nameLen;
            unsigned j;

            fseek(f,e.lfo+baseOfs+baseSrcFile[i],SEEK_SET);
            fread(&table,1,sizeof(table),f);
            fread(baseSrcLin,4,table.cSeg,f);
            fread(se,8,table.cSeg,f);
            fread(&nameLen,1,1,f);

            // add new source file
            assert(nrSrc<sizeSource);
            src[nrSrc].name=strBuf+curStrBuf;
            src[nrSrc].from=0;
            src[nrSrc].to=0;
            src[nrSrc].nrLine=0;
            src[nrSrc].line=line+curLine;
            fread(strBuf+curStrBuf,1,nameLen,f);
            curStrBuf+=nameLen+1;

            for (j=0;j<table.cSeg;j++) 
            {
              tLineHeader hdr;

              fseek(f,e.lfo+baseOfs+baseSrcLin[j],SEEK_SET);
              fread(&hdr,sizeof(hdr),1,f);
              if (hdr.Seg==1) 
              {
                unsigned l;

                // adjust from/to size
                if (!src[nrSrc].to)
                {
                  src[nrSrc].from=se[j].start;
                  src[nrSrc].to=se[j].end+1;
                }
                else
                {
                  if (se[j].start<src[nrSrc].from)
                    src[nrSrc].from=se[j].start;
                  if (se[j].end>=src[nrSrc].to)
                    src[nrSrc].to=se[j].end+1;
                }

                // read line numbers
                assert(curLine<sizeLine);
                for (l=0;l<hdr.cPair;l++)
                  fread(&line[curLine+l].ofs,4,1,f);
                for (l=0;l<hdr.cPair;l++)
                  fread(&line[curLine+l].n,2,1,f);
                curLine+=hdr.cPair;
                src[nrSrc].nrLine+=hdr.cPair;
              }
            }
            nrSrc++;
          }
          fseek(f,save,SEEK_SET);
        }
        break;
    }
  }

  return 1;
}

#define GET_OFFSET(strc,member) ((unsigned)(&(((strc*)0)->member)))
static int sortOffset;
#define _UNSIGNED(x)  (*(unsigned *)(((char *)(x))+sortOffset))

static int compare(const void *a, const void *b)
{
  if (_UNSIGNED(a)<_UNSIGNED(b)) return -1;
  if (_UNSIGNED(a)>_UNSIGNED(b)) return 1;
  return 0;
}

static int search(const void *a, const void *b)
{
  if (*(unsigned *)a<_UNSIGNED(b)) return -1;
  if (*(unsigned *)a>_UNSIGNED(b)) return 1;
  return 0;
}

static void *_bsearch(const void *key, const void *base, size_t num, size_t width, 
                      int (*compare ) ( const void *elem1, const void *elem2 ) )
{
  char *cur=(char *)base;

  if (!num) return NULL;

  while (num>0) 
  {
    if (compare(key,cur)<0) return cur==base?cur:cur-width;
    cur+=width;
    num--;
  }

  return cur-width;
}

static void sortData(void)
// sorts data: proc by increasing from-offset, src by increasing from-offset,
// lines by increasing from-offset within src context
// seeks matching source files for proc entries
{
  int k;

  // sort first
  sortOffset=GET_OFFSET(tProc,from);
  qsort(proc,nrProc,sizeof(tProc),compare);

  sortOffset=GET_OFFSET(tSource,from);
  qsort(src,nrSrc,sizeof(tSource),compare);

  sortOffset=GET_OFFSET(tLine,ofs);
  for (k=0;k<nrSrc;k++)
    qsort(src[k].line,src[k].nrLine,sizeof(tLine),compare);

  // now try to match
  sortOffset=GET_OFFSET(tSource,from);
  for (k=0;k<nrProc;k++)
  {
    tSource *s=(tSource *)_bsearch(&proc[k].from,src,nrSrc,sizeof(tSource),search);
    if (s&&s->from<=proc[k].from&&
        s->to>=proc[k].to)
      proc[k].src=s;
  }
}

int cvLoad(char *fn)
// load codeview info from given file, ==0 on error, !=0 if successful
// fn=NULL: load current module
{
  char buf[4];
  unsigned relOfs;
  unsigned compare;

  // get own EXE filename
  if (!fn)
  {
    static char namebuf[256];
    char *help;

    strncpy(namebuf,GetCommandLine(),255);
    namebuf[255]=0;

    help=namebuf;
    if (*help=='\"') 
    {
      // quoted filename
      help++;
      while (*help&&*help!='\"') help++;
      *help=0;
      fn=namebuf+1;
    }
    else
    {
      // regular filename
      while (*help&&*help!=' ') help++;
      *help=0;
      fn=namebuf;
    }
  }

  // open file
  f=fopen(fn,"rb");
  if (!f) ERR("Can't open file");

  // ID ok?
  fseek(f,-8,SEEK_END);
  fread(buf,1,4,f);
  if (buf[0]!='N'||buf[1]!='B') 
  {
    fclose(f); ERR("No CV info");
  }

  // get base offset
  fread(&relOfs,4,1,f);
  baseOfs=ftell(f)-relOfs;

  // offset ok?
  fseek(f,baseOfs,SEEK_SET);
  fread(&compare,4,1,f);
  if (compare!=*(unsigned *)buf) 
  {
    fclose(f); ERR("Wrong ID");
  }

  // get directory offset
  fread(&dirStart,4,1,f);
  dirStart+=baseOfs;

  // calc size of memory to be allocated
  // and allocated structures
  if (!calcNeededMemory()) 
  {
    fclose(f); ERR("Memory calculation failed");
  }

  // load data into memory
  if (!loadData())
  {
    fclose(f); ERR("Loading failed");
  }
  fclose(f);

  // sort data
  sortData();

  return 1;
}

int cvQuery(unsigned ofs, tCVQuery *q)
// searches for info for given offset, !=0 found info, ==0 no info
{
  tProc *p;

  sortOffset=GET_OFFSET(tProc,from);
  p=(tProc *)_bsearch(&ofs,proc,nrProc,sizeof(tProc),search);
  if (!p) return 0;

  if (p->to<ofs) return 0;
  memset(q,0,sizeof(tCVQuery));
  q->procName=p->name;
  q->procFrom=p->from;
  q->procTo=p->to;

  if (p->src)
  {
    tLine *l;

    q->srcName=p->src->name;
    q->srcFrom=p->src->from;
    q->srcTo=p->src->to;
    
    sortOffset=GET_OFFSET(tLine,ofs);
    l=(tLine *)_bsearch(&ofs,p->src->line,p->src->nrLine,sizeof(tLine),search);
    if (l) 
    {
      q->lineOfs=l->ofs;
      q->line=l->n;
    }
  }

  return 1;
}
